# Copyright (c) 2020 Paul Barker <pbarker@konsulko.com>
# SPDX-License-Identifier: Apache-2.0

import yaml


from . import agents


_CONFIG = {}


DEFAULT_OPTIONS = {"dest": ""}


class Source:
    def __init__(self, name, properties):
        self.properties = properties
        self.properties["name"] = name
        self.agent = agents.get(properties)

    def mirror(self):
        return self.agent.mirror()


class Config:
    def __init__(self, cfg):
        self.cfg = cfg

    def sources(self):
        cfg_sources = self.cfg.get("sources", [])
        return [Source(name, properties) for name, properties in cfg_sources.items()]

    def option(self, name):
        options = self.cfg.get("options", {})
        default_value = DEFAULT_OPTIONS.get(name)
        return options.get(name, default_value)


def load(filename):
    global _CONFIG
    f = open(filename, "r")
    _CONFIG = Config(yaml.safe_load(f))
    return _CONFIG


def get():
    return _CONFIG



# Copyright (c) 2020 Paul Barker <pbarker@konsulko.com>
# SPDX-License-Identifier: Apache-2.0


class Agent:
    def __init__(self, properties):
        self.properties = properties

    def mirror(self):
        raise NotImplementedError



# Copyright (c) 2020 Paul Barker <pbarker@konsulko.com>
# SPDX-License-Identifier: Apache-2.0

from .command import Command
from .git import Git
from .gitlab import Gitlab
from .rclone import RClone


agents = {
    "command": Command,
    "gitlab": Gitlab,
    "git": Git,
    "rclone": RClone,
}


def get(properties):
    # The default connection method for a source is the name of the source
    # itself
    agent_name = properties.get("agent", properties.get("name"))
    agent_ctor = agents[agent_name]
    return agent_ctor(properties)



# Copyright (c) 2020 Paul Barker <pbarker@konsulko.com>
# SPDX-License-Identifier: Apache-2.0

import os
import subprocess
from urllib.parse import urljoin


from .. import config
from .base import Agent


def do_fetch(url, local_path):
    print(f"Fetching in {local_path}")
    subprocess.run(["git", "fetch", "--all", "-q"], cwd=local_path, check=True)


def do_clone(url, local_path):
    print(f"Cloning into {local_path}")
    parent_path = os.path.dirname(local_path)
    os.makedirs(parent_path, exist_ok=True)
    subprocess.run(["git", "clone", "--mirror", "-q", url, local_path], check=True)


class Git(Agent):
    def mirror(self):
        name = self.properties.get("name")
        url_prefix = self.properties.get("url_prefix", "")
        repositories = self.properties.get("repositories")
        dest = config.get().option("dest")

        for repo in repositories:
            url = urljoin(url_prefix, repo)
            local_path = os.path.join(dest, name, repo)
            if not local_path.endswith(".git"):
                local_path += ".git"

            if os.path.exists(local_path):
                do_fetch(url, local_path)
            else:
                do_clone(url, local_path)



from invoke import task

@task
def install(c):
    """Install the project locally"""
    c.run("pip install .")

@task
def build(c):
    """Build the project"""
    c.run("python3 -m pep517.build .")

@task
def clean(c):
    """Remove build output"""
    c.run("rm -rf build dist src/*.egg-info")

@task
def release(c, version):
    """Release a version of the project"""
    clean(c)
    c.run(f"sed -i 's/^\(__version__ =\).*$/\\1 \"{version}\"/' src/mirrorshades/__init__.py")
    c.run(f"git commit -s -m 'Release {version}' src/mirrorshades/__init__.py")
    build(c)
    c.run(f"echo mirrorshades {version} > dist/RELEASE_NOTES.txt")
    c.run(f"markdown-extract -fri '^{version}' ChangeLog.md >> dist/RELEASE_NOTES.txt")
    c.run(f"git tag -a -F dist/RELEASE_NOTES.txt 'v{version}' HEAD")
    with c.cd("dist"):
        c.run("sha256sum * > SHA256SUMS")



# Copyright (c) 2020 Paul Barker <pbarker@konsulko.com>
# SPDX-License-Identifier: Apache-2.0

import argparse


from . import config


__version__ = "0.1.3"


def parse_args():
    parser = argparse.ArgumentParser(
        prog="mirrorshades", description="Data mirroring tool"
    )
    parser.add_argument(
        "config_path",
        default="mirrorshades.yml",
        nargs="?",
        help="path to the configuration file (defaults to 'mirrorshades.yml' "
        "in the current directory)",
    )
    parser.add_argument(
        "--version", action="version", version=f"mirrorshades {__version__}"
    )
    return parser.parse_args()


def main():
    args = parse_args()
    cfg = config.load(args.config_path)

    for source in cfg.sources():
        source.mirror()



# Copyright (c) 2020 Paul Barker <pbarker@konsulko.com>
# SPDX-License-Identifier: Apache-2.0

import os
import subprocess


from .. import config
from .base import Agent


class RClone(Agent):
    def mirror(self):
        name = self.properties.get("name")
        remote = self.properties.get("remote")
        paths = self.properties.get("paths")
        dest = config.get().option("dest")

        for path in paths:
            if path == ".":
                path = ""
            full_path = ":".join([remote, path])
            local_path = os.path.join(dest, name, path)
            subprocess.run(
                ["rclone", "sync", "--create-empty-src-dirs", full_path, local_path],
                check=True,
            )



# Copyright (c) 2020 Paul Barker <pbarker@konsulko.com>
# SPDX-License-Identifier: Apache-2.0

import subprocess


from .. import config
from .base import Agent


class Command(Agent):
    def mirror(self):
        command = self.properties.get("command")
        subprocess.run(command, shell=True, check=True)



# Copyright (c) 2020 Paul Barker <pbarker@konsulko.com>
# SPDX-License-Identifier: Apache-2.0

import sys
from urllib.parse import urlparse


from .base import Agent
from .git import Git


class Gitlab(Agent):
    def mirror(self):
        try:
            import gitlab
        except ModuleNotFoundError:
            print(
                "The `gitlab` python package is needed to mirror from GitLab.",
                file=sys.stderr,
            )
            print(
                "Please install this (for example, using `pip install gitlab`) and try again.",
                file=sys.stderr,
            )
            sys.exit(1)

        name = self.properties.get("name")
        private_token = self.properties.get("private_token")
        server = self.properties.get("server", "https://gitlab.com")
        groups = self.properties.get("groups", [])
        projects = self.properties.get("projects", [])
        repositories = []

        gl = gitlab.Gitlab(server, private_token=private_token)

        for group_path in groups:
            group = gl.groups.get(group_path)
            for project in group.projects.list(include_subgroups=True):
                repositories.append(project.path_with_namespace)

        for project_path in projects:
            project = gl.projects.get(project_path)
            repositories.append(project.path_with_namespace)

        url = urlparse(server)
        location_with_auth = f"oauth2:{private_token}@{url.netloc}"
        url_prefix = url._replace(netloc=location_with_auth).geturl()
        git_properties = {
            "name": name,
            "url_prefix": url_prefix,
            "repositories": repositories,
        }
        git = Git(git_properties)
        git.mirror()
